/* write one line from stdin to the file specified on the command line.
   This program uses separation of privilege. */

#include <sys/param.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <err.h>
#include <pwd.h>
#include <errno.h>
#include <unistd.h>

const char *unpriv_user = "nobody";
const char *basedir = "/tmp/root";

void string_validate(const char *fname)
{
    if (strlen(fname) == 0)
        errx(1, "unpriv: 0-length file name");
    if (strlen(fname) > MAXPATHLEN)
        errx(1, "unpriv: path exceeds %d characters", MAXPATHLEN);
    if (fname[0] != '/')
        errx(1, "unpriv: file name must be an absolute path");
}

/* wrapper around realloc to avoid integer overflow */
void *xrealloc(void *ptr, const size_t nelem, const size_t size)
{
    /* BUG: the check below can be fooled; for extra credit, figure out how */
    if (nelem * size < size)
	errx(1, "xrealloc: size overflow (%ld * %ld = %ld)", nelem,
	     size, size * nelem);
    return realloc(ptr, nelem * size);
}

char *readline(FILE *f)
{
    int c;
    char *line;
    unsigned int size, used;
    const unsigned int grow = 10;

    line = calloc(grow, sizeof(char));
    if (line == NULL)
        err(1, "readline: calloc failed");
    size = grow;
    used = 0;

    while ((c = getc(f)) != EOF && c != '\n') {
        if (used+1 > size-1) {
	    /* Check for overflow */
	    if (size + grow < size)
	        errx(1, "readline: size overflow (%d + %d = %d)", size,
		     grow, size + grow);
	    size += grow;

	    line = xrealloc(line, size, sizeof(char));
	    if (line == NULL)
	        err(1, "readline: realloc failed");
	}
	
	line[used++] = c;
    }
    if (c == EOF)
        return NULL;

    line[++used] = '\0'; /* space for the null is guaranteed in loop */

    return line;
}

void drop_privs(void)
{
    struct passwd *new;

    errno = 0;
    if ((new = getpwnam(unpriv_user)) == NULL) {
        if (errno == 0)
            errx(1, "drop_privs: User '%s' not in passwd file\n",
                 unpriv_user);
        else
            err(1, "drop_privs: getpwnam(%s) failed", unpriv_user);
    }
    endpwent();

    if (setgid(new->pw_gid) < 0)
        err(1, "drop_privs: set gid to %d failed", new->pw_gid);

    if (setuid(new->pw_uid) < 0)
        err(1, "drop_privs: set uid to %d failed", new->pw_uid);
}

void unpriv_ops(const char *fname, const int priv_pipe)
{
    FILE *priv;
    char *line;

    if ((priv = fdopen(priv_pipe, "w")) == NULL)
        err(1, "Unpriv: fdopen of pipe to priv part failed");

    /* Cannot validate file name other than simple string checks */
    string_validate(fname);

    line = readline(stdin);
    if (line == NULL || line[0] == '\0')
        errx(1, "Unpriv: Empty line; operation cancelled");

    fprintf(priv, "%s\n", fname);
    fprintf(priv, "%s\n", line);

    if (fclose(priv) == EOF)
        err(1, "Unpriv: fclose of pipe to priv failed");

    printf("Unpriv: wrote two lines:\n%s\n%s\n", fname, line);

    free(line);
}

void priv_ops(const int unpriv_pipe)
{
    FILE *unpriv;
    char *fname, *line;
    unsigned int i, len;
    int rc, fd;

    if ((unpriv = fdopen(unpriv_pipe, "r")) == NULL)
        err(1, "Priv: fdopen of pipe from unpriv part failed");

    fname = readline(unpriv);
    if (fname == NULL || fname[0] == '\0')
        errx(1, "priv: empty fname; operation cancelled");
    line = readline(unpriv);
    if (line == NULL || line[0] == '\0')
        errx(1, "priv: empty line; operation cancelled");

    /* validate file name */
    string_validate(fname);
    for (i=0; basedir[i] != '\0' && fname[i] != '\0'; i++)
        if (basedir[i] != fname[i])
	    errx(1, "priv: %s is not within %s", fname, basedir);
    if (i < strlen(basedir))
	errx(1, "priv: %s is not within %s", fname, basedir);

    if ((fd = open(fname, O_WRONLY|O_CREAT|O_EXCL, 0644)) == -1)
        err(1, "priv: exclusive open of '%s' failed", fname);

    /* do the write */
    /* Should factor out the common code and make a function */
    len = strlen(line);
    if ((rc = write(fd, line, len)) != len) {
	if (rc < 0)
	    err(1, "priv: write line failed");
	else
	    errx(1, "priv: incomplete write");
    }
    if ((rc = write(fd, "\n", 1)) != 1) {
	if (rc < 0)
	    err(1, "priv: write newline failed");
	else
	    errx(1, "priv: incomplete second write");
    }

    if (close(fd) < 0)
        err(1, "priv: close '%s' failed", fname);

    free(fname);
    free(line);
}

/* set up a pipe between parent and child.  Child retains
   privs, parent drops them.  Call functions to do proper
   operations for each process. */
void do_privsep(const char *fname)
{
    int pid, cpid, pipefd[2], status;

    if (pipe(pipefd) < 0)
        err(1, "pipe() failed");

    switch(pid = fork()) {
        case -1: /* error */
            err(1, "Fork failed");
            break; /* NOTREACHED */
        case 0: /* child; read from pipe, keep privs */
            if (close(pipefd[1]) < 0)
                err(1, "close writing end of pipe failed");
            priv_ops(pipefd[0]);
            break;
        default: /* parent; drop privs, write to pipe */
            if (close(pipefd[0]) < 0)
                err(1, "close reading end of pipe failed");
            drop_privs();
            unpriv_ops(fname, pipefd[1]);
   
            printf("Unpriv: about to wait for priv child\n");
            cpid = wait(&status);
            if (cpid < 0)
                err(1, "wait failed");
            if (pid != cpid)
                errx(1, "Weird, child pid %d != wait pid %d",
                     cpid, pid);
            printf("Unpriv: priv part finished, exit status %d\n",
                   WEXITSTATUS(status));
            break;
    }
}

int main(int argc, char *argv[])
{
    if (argc != 2)
        errx(1, "Usage: %s path\npath is absolute", argv[0]);

    do_privsep(argv[1]);

    return 0;
}
